#!/usr/local/bin/perl
# 
# Usage :
#
# flspec sourcelistset="Name of the fits file containing the list of 
#                     observed sources" 
#        set="EVENT LIST filename of the EPIC imaging observation"
#        imageset="EPIC Raw Image in band " 
#        exposureset="EPIC Exposure Map corresponding to image"
#        tmpflatset="Name of the file containing the temporary 
#                     flatfielded image"
#        tmpbkgregionset="Name of the temporary file containing the regions 
#                     for the sources background"
#        tmpposmaskset="Name of the file containing the temporary positive 
#                      image mask"
#        tmpnegmaskset="Name of the file containing the temporary negative 
#                      image mask"
#        tmpposspecset="Name of the file containing a temporary positive 
#                      spectrum"
#        tmpnegspecset="Name of the file containing a temporary negative 
#                      spectrum"
#        tmpdiffset="Name of the file containing a temporary subtracted
#                      spectrum"
#        tmpsrclistset="Name of the temporary file used by region during
#               filtering of source list" 
#        binsize="Size of the rebinned pixels in degrees to define 
#                 positive and negative fluctuations"
#        spectrumset="Output fits filename containing the spectrum of the 
#               background fluctuations at different offaxis angles" 
#     
#     
#
# ------------------------------------------------------
# Define variables   
# ------------------------------------------------------ 

use SAS;
use strict;

sub flspec{
  
  my $binsize;                # deg for binned image pixel size
  my $rmax=12.0;              # arcm : maximum radius for annuli
  my $nann=10;                # number of annuli
  my $XcolMask="X";           # columns to extract spectra
  my $YcolMask="Y";           # 
  my $specBinMOS=15;
  my $specBinPN=5;
  my $specChanMaxMOS=11999;
  my $specChanMaxPN=20479;
  my ($att,$attCol);
  my ($coX,$coXPseudo,$coY,$coYPseudo,$coPI);
  my ($evPseudo,$evselBinSizePar,$evtlist,$exist,$existExposDs,$existExposTb);
  my ($expm,$expmContent,$exposds,$expostb,$expr,$extname);
  my ($factor,$factorFormat,$filter,$filterexp,$flatimg,$flspecfile);
  my ($hdu,$ids,$img,$imgCkeys,$imgContent,$imgPrimary,$inPixSize,$inst);
  my ($jj,$jj1,$key,$label,$msg,$nhdus,$negat_mask,$nspec); 
  my ($off_arcs,$pattern,$posit_mask,$radius_int,$rowidStr);
  my ($SolidAngle,$SolidAngleFormat,$SolidAngle_neg,$SolidAngle_pos,$specBin);
  my ($specChanMax,$spec_neg,$spec_pos,$srclist);
  my ($temp1,$temp2,$temp3,$tcdltX,$verb,$xColPixSize);
  my @dset=();
  my @inimgdim=();
  my @radius_ext;
  my @off_arcm;
  my @pseudoStrAtt=("FILTER","INSTRUME","TELESCOP","AVRG_PNT","RADECSYS","REFYCUNI","REFXCUNI","REFYCTYP");
  my @pseudoR32Att=("PA_PNT","RA_PNT","DEC_PNT");
  push(@pseudoR32Att,"EQUINOX","RA_NOM","DEC_NOM","RA_OBJ","DEC_OBJ");
  push(@pseudoR32Att,"REFYCUNI","REFYDMAX","REFYDMIN","REFYLMAX","REFYLMIN");
  push(@pseudoR32Att,"REFYCDLT","REFYCRVL","REFYCRPX","REFYCTYP","REFXCUNI");
  push(@pseudoR32Att,"REFXDMAX","REFXDMIN","REFXLMAX","REFXLMIN");
  my @pseudoStrAttCol=("TCUNI","TCTYP");
  my @pseudoR32AttCol=("TCDLT","TCRVL","TCRPX","TDMAX","TDMIN","TLMAX","TLMIN");

# ------------------------------------------------------ 
# Read input parameters
# ------------------------------------------------------ 

  $srclist=stringParameter("sourcelistset");
  $evtlist=stringParameter("set");
  $img=stringParameter("imageset");
  $expm=stringParameter("exposureset");
  $flatimg=stringParameter("tmpflatset");
  $temp1=stringParameter("tmpbkgregionset");
  $temp2=stringParameter("tmpsrclistset");
  $posit_mask=stringParameter("tmpposmaskset");
  $negat_mask=stringParameter("tmpnegmaskset");
  $spec_pos=stringParameter("tmpposspecset");
  $spec_neg=stringParameter("tmpnegspecset");
  $temp3=stringParameter("tmpdiffset");
  $binsize=realParameter("binsize");
  $flspecfile=stringParameter("spectrumset");

#--------------------------------------------------------
# Find out TCRPX, TCDLT column (X,Y) attributes
#--------------------------------------------------------
 

  # first get column numbers (others required later on)
  if(system("fcolpar infile=$evtlist\[EVENTS\] colname='X'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFcolpar","Cannot establish column number for X in eventlist\n");
  }
  $coX=`pget fcolpar colnum`;
  chomp($coX);
  if(system("fcolpar infile=$evtlist\[EVENTS\] colname='Y'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFcolpar","Cannot establish column number for Y in eventlist\n");
  }
  $coY=`pget fcolpar colnum`;
  chomp($coY);

  if(system("fcolpar infile=$evtlist\[EVENTS\] colname='PI'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFcolpar","Cannot establish column number for PI in eventlist\n");
  }
  $coPI=`pget fcolpar colnum`;
  chomp($coPI);

  # ---------------------------------

  $att="TCDLT".$coX;
  if(system("fkeypar fits=$evtlist\[EVENTS\] key=$att")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFkeypar","Cannot run fkeypar to read $att from eventlist");
  }
  $tcdltX=`pget fkeypar value`;
  chomp($tcdltX);
  $xColPixSize=abs($tcdltX);



# ------------------------------------------------------
#     Check : input FITS files are IMAGE/EXPOSURE MAP
# ------------------------------------------------------

  $key=" ";
  if(system("fstruct $expm >/dev/null")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFstruct","Cannot run fstruct for \"$expm\"\n");
  }
  ($nhdus=`pget fstruct totalhdu`) || ($nhdus = 1);

  for ($hdu=0;$hdu<$nhdus;$hdu++){
    $expmContent=$expm."+".$hdu;
    if(system("fkeypar fits=$expmContent key='CONTENT'")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFkeypar","Cannot run fkeypar to read CONTENT from \"$expmContent\n");
    }
    $key=`pget fkeypar value`;
    chomp($key);
    ($key =~ /EPIC\s*EXPOSURE\s*MAP\s*/) && last;
  }
  unless ($key =~ /EPIC\s*EXPOSURE\s*MAP\s*/){
    SAS::error("NotExpMap","Input image:\"$expm\", is not an EPIC EXPOSURE MAP file\n");
  }


  $key=" ";
  if(system("fstruct $img >/dev/null")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFstruct","Cannot run fstruct for \"$img\"\n");
  }
  ($nhdus=`pget fstruct totalhdu`) || ($nhdus = 1);
  for ($hdu=0;$hdu<$nhdus;$hdu++){
    $imgContent=$img."+".$hdu;
    if(system("fkeypar fits=$imgContent key='CONTENT'")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFkeypar","Cannot run fkeypar to read CONTENT from image\n");
    }
    $key=`pget fkeypar value`;
    chomp($key);
    ($key =~ /EPIC\s*IMAGE/) && last;
  }
  unless($key =~ /EPIC\s*IMAGE/){
    SAS::error("NotImg","Input image:\"$img\", is not an EPIC IMAGE file\n");
  }


# ------------------------------------------------------
# Obtain some pixel size/position WCS attributes from imageset
# ------------------------------------------------------

  $imgCkeys=$img."+0";
  if(system("fkeypar fits=$imgCkeys key='CDELT1'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFkeypar","Cannot run fkeypar to read CDELT1 from image\n");
  }
  $exist=`pget fkeypar exist`;
  unless($exist =~ /yes/){
    $imgCkeys=$img."+1";
    if(system("fkeypar fits=$imgCkeys key='CDELT1'")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFkeypar","Cannot run fkeypar to read CDELT1 from image\n");
    }
  }
  $inPixSize=`pget fkeypar value`;
  chomp($inPixSize);
  $inPixSize=abs($inPixSize);


# ------------------------------------------------------
#     Calculate INPUT IMAGE dimensions
# ------------------------------------------------------

  $imgPrimary=$img."\[PRIMARY\]";
  if(system("fkeypar fits=$imgPrimary key='NAXIS1'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFkeypar","Cannot run fkeypar to read NAXIS1 from image\n");
  }
  $inimgdim[0]=`pget fkeypar value`;
  chomp($inimgdim[0]);
  if(system("fkeypar fits=$imgPrimary key='NAXIS2'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFkeypar","Cannot run fkeypar to read NAXIS2 from image\n");
  }
  $inimgdim[1]=`pget fkeypar value`;
  chomp($inimgdim[1]);
  $inimgdim[0] =~ s/\s+//g;
  $inimgdim[1] =~ s/\s+//g;

# ------------------------------------------------------
# FLUCTUATIONS CALCULUS...
# ------------------------------------------------------
  # binSize parameter for evselect when making images
  $evselBinSizePar=($inPixSize/$xColPixSize);
  $evselBinSizePar=sprintf("%.0f",$evselBinSizePar);

#
# ------------------------------------------------------
# 1) Write background global region removing ALL sources in SRCLIST
#    through task: REGION
#    Input:  SRC_LIST_file
#
#    Output: BKG_REGION_file --> temp1
#            temp_file       --> temp2 (not kept)
# ------------------------------------------------------
#
  $srclist=$srclist.":SRCLIST";

  $msg="Writing regions for all sources in source list...\n";
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

  if(system("region eventset=$evtlist srclisttab=$srclist operationstyle='global' bkgregionset=$temp1 radiusstyle='enfrac' energyfraction=0.95 outunit='xy' tempset=$temp2 expression='ID_BAND==2'")){

    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrRegion","Cannot run REGION on \"$srclist\"\n");
  }
  (-e $temp2) && unlink($temp2); # rm REGION temp file


# ------------------------------------------------------ 
#  2) Create a mask to "hide" source regions in input image
#---------------------------------------------------------
#
#     2.1) Create pseudo evt list with X/Y/DETX/DETY/THETA columns
#          (run createpseudoevt F90 module) --> temp2
#

  if(system("createpseudoevt set=$temp2 imageset=$img inset=$evtlist")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrPseudo","Cannot create Pseudo EVT");
  }
  $evPseudo=$temp2; # easier to remember

#
#     2.2) Write WCS keywords (from input EVT list) to PSEUDO event list
#
#
#          a) locate columns X/Y numbers in Pseudo EVT
#

  if(system("fcolpar infile=$evPseudo\[EVENTS\] colname='X'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFcolpar","Cannot establish column number for X in Pseudo eventlist\n");
  }
  $coXPseudo=`pget fcolpar colnum`;
  chomp($coXPseudo);
  if(system("fcolpar infile=$evPseudo\[EVENTS\] colname='Y'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFcolpar","Cannot establish column number for Y in Pseudo eventlist\n");
  }
  $coYPseudo=`pget fcolpar colnum`;
  chomp($coYPseudo);

#          b) Add String Type Attributes to Pseudo EVT list

  foreach $att (@pseudoStrAtt){
    if(system("fkeypar fits=$evtlist\[EVENTS\] key=$att")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFkeypar","Cannot run fkeypar to read $att from eventlist\n");
    }
    $key=`pget fkeypar value`;
    chomp($key);
    if(system("fparkey fits=$evPseudo\[EVENTS\] add=yes value=$key key=$att")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFparkey","Cannot run fparkey to add $att to Pseudo event list\n");
    }

#          c) define 'standard' binning of spectra for MOS/PN

    if($att =~ /INSTRUME/){
      if($key =~ /EMOS/){
	$specBin=$specBinMOS;
	$specChanMax=$specChanMaxMOS;
      }elsif($key =~ /EPN/){
	$specBin=$specBinPN;
	$specChanMax=$specChanMaxPN;
      }
      $inst=$key;
    }
  }

#          d) Add Real32 attributes

  foreach $att (@pseudoR32Att){
    if(system("fkeypar fits=$evtlist\[EVENTS\] key=$att")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFkeypar","Cannot run fkeypar to read $att from eventlist\n");
    }
    $key=`pget fkeypar value`;
    chomp($key);
    if(system("fparkey add=yes fits=$evPseudo\[EVENTS\] value=$key key=$att")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFparkey","Cannot run fparkey to add $att to Pseudo event list\n");
    }
  }

#          e) Add String Column attributes

  foreach $att (@pseudoStrAttCol){
    $attCol=$att."$coX";
    if(system("fkeypar fits=$evtlist\[EVENTS\] key=$attCol")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFkeypar","Cannot run fkeypar to read $attCol from eventlist\n");
    }
    $key=`pget fkeypar value`;
    chomp($key);
    $attCol=$att."$coXPseudo";
    if(system("fparkey add=yes fits=$evPseudo\[EVENTS\] value=$key key=$attCol")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFparkey","Cannot run fparkey to add $attCol to Pseudo event list\n");
    }

    $attCol=$att."$coY";
    if(system("fkeypar fits=$evtlist\[EVENTS\] key=$attCol")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFkeypar","Cannot run fkeypar to read $attCol from eventlist\n");
    }
    $key=`pget fkeypar value`;
    chomp($key);
    $attCol=$att."$coYPseudo";
    if(system("fparkey add=yes fits=$evPseudo\[EVENTS\] value=$key key=$attCol")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFparkey","Cannot run fparkey to add $attCol to Pseudo event list\n");
    }
  }

#          f) Add Real32 Column Attributes

  foreach $att (@pseudoR32AttCol){
    $attCol=$att."$coX";
    if(system("fkeypar fits=$evtlist\[EVENTS\] key=$attCol")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFkeypar","Cannot run fkeypar to read $attCol from eventlist\n");
    }
    $key=`pget fkeypar value`;
    chomp($key);

    $attCol=$att."$coXPseudo";
    if(system("fparkey add=yes fits=$evPseudo\[EVENTS\] value=$key key=$attCol")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFparkey","Cannot run fparkey to add $attCol to Pseudo event list\n");
    }


    $attCol=$att."$coY";
    if(system("fkeypar fits=$evtlist\[EVENTS\] key=$attCol")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFkeypar","Cannot run fkeypar to read $attCol from eventlist\n");
    }
    $key=`pget fkeypar value`;
    chomp($key);

    $attCol=$att."$coYPseudo";
    if(system("fparkey add=yes fits=$evPseudo\[EVENTS\] value=$key key=$attCol")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFparkey","Cannot run fparkey to add $attCol to Pseudo event list\n");
    }
  }


#
#          g) XMM attributes so that evselect does not complain
#

  if(system("fparkey add=yes value='XMM' fits=$evPseudo+0 key=TELESCOP")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFparkey","Cannot run fparkey to add TELESCOP to Pseudo event list\n");
  }
  if(system("fparkey add=yes value='XMM' fits=$evPseudo+1 key=TELESCOP")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFparkey","Cannot run fparkey to add TELESCOP to Pseudo event list\n");
  }
  if(system("fparkey add=yes value=$inst fits=$evPseudo+0 key=INSTRUME")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFparkey","Cannot run fparkey to add INSTRUME to Pseudo event list\n");
  }
  if(system("fparkey add=yes value=$inst fits=$evPseudo+1 key=INSTRUME")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFparkey","Cannot run fparkey to add INSTRUME to Pseudo event list\n");
  }

#
#     2.3) Filter pseudo evt list with region output (temp1) and save
#          image (=mask) ---> temp3 (same size/dimensions as input image)
#          Mask: "0" in source position and "1" otherwise

  if(system("evselect table=$evPseudo:EVENTS  expression='region($temp1,X,Y)' withfilteredset=false keepfilteroutput=false destruct=false writedss=false updateexposure=false withimageset=true imageset=$temp3 xcolumn='X' ycolumn='Y' imagebinning=binSize ximagebinsize=$evselBinSizePar yimagebinsize=$evselBinSizePar")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrEvselect","Cannot run EVSELECT to filter input evt list with REGION output\n");
  }
#
# ------------------------------------------------------ 
#  3) Flatfield input IMAGE with its EXPOSMAP and mask out the sources
#       Input:  IMG_file
#               EXPOSMAP_file
#
#       Output: FLAT_IMG_file
# ------------------------------------------------------ 
#
  #
  #   3.0) create mask to "clean" borders of expmap
  #        flatimg is a temp file here
  #        final expmap in temp1
  if(system("emask expimageset=$expm detmaskset=$flatimg threshold1=0.1")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrEmask","Cannot create mask for \"$expm\"\n");
  }
  if(system("farith infil1=$expm infil2=$flatimg outfil=$temp1 ops=MUL datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFarith","Cannot multiply: exposure_mask and exposure_map\n");
  }
  (-e $flatimg) &&  unlink($flatimg);

  #
  #   3.1) Calculate imageset/exposuremap_masked= flatimg
  #

  $msg="Flatfielding input image with its corresponding exposure map...\n";
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

  if(system("farith infil1=$img infil2=$temp1 outfil=$flatimg ops=DIV datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFarith","Cannot divide:\"$img\" and \"$expm\"\n");
  }
  #
  #   3.2) flatimg/region mask (to hide sources) --> temp1
  #                  (temp3)
  #      Resulting image: "-1" in source positions and flatimg values otherwise
  #

  if(system("farith infil1=$flatimg infil2=$temp3 outfil=$temp1 ops=DIV datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFarith","Cannot calculate flatfield/regionmask\n");
  }
  unlink($flatimg);
  unlink($temp3);

#  Once input image has been flatfielded and sources hidden (temp1):
# ------------------------------------------------------------------
#  4) Create "median" mask file to select pixels above/below
#     the "median": FLMASK (F90 module program)
#
#      Input:  FLAT_IMG_file
#
#      Output: MASK_POS_file
#              MASK_NEG_file
# -------------------------------------------------------------------
#

  $msg="Creating mask files...\n";
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

  if(system("flmask set=$evtlist imageset=$temp1 binsize=$binsize positivemaskset=$posit_mask negativemaskset=$negat_mask")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFlmask","Mask files cannot be created\n");
  }
  unlink($temp1);

#
# ------------------------------------------------------
#  6) Define annular regions for background spectra.
# ------------------------------------------------------
#
  for($jj=0;$jj<$nann;$jj++){
    $jj1=$jj+1;
    $msg="Spatial region number: $jj1\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

    $radius_ext[$jj]=sqrt(($jj+1)/$nann)*$rmax * 60.0; # arcs

    if($jj==0){
      $filter="'"."THETA <= $radius_ext[$jj]"."'";
      $off_arcs=$radius_ext[$jj]/2.0;
      $radius_int=0.0;
    }else{
      $radius_int=$radius_ext[$jj-1];
      $filter="'"."THETA > $radius_int && THETA <= $radius_ext[$jj]"."'";
      $off_arcs=($radius_ext[$jj]+$radius_ext[$jj-1])/2.0;
    }
    $off_arcm[$jj]=$off_arcs/60.0;


#-------------------------------------------------------------
#    7) Filter pseudo evt list to obtain annulus_mask:$temp1
#-------------------------------------------------------------


    if(system("evselect table=$evPseudo:EVENTS  expression=$filter withfilteredset=false keepfilteroutput=false destruct=false writedss=false updateexposure=false withimageset=true imageset=$temp1 xcolumn='X' ycolumn='Y' imagebinning=binSize ximagebinsize=$evselBinSizePar yimagebinsize=$evselBinSizePar")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrEvselect","Cannot filter Pseudo EVT to obtain annulusmask\n");
    }
#
# ------------------------------------------------------
#  8) Filter EVT_LIST with POS/NEG MASKs * annular mask
#     and extract "positive/negative" spectra : EVSELECT
#     Input:  EVT_LIST_file
#             MASK_POS(NEG)_mask*ANNULAR_mask
#
#     Output: SPEC_POS(NEG)_file_in_annulus
# ------------------------------------------------------
#
	
#
#     8.1) Do annulus_mask(temp1)*posit_mask(+1)=final posit_mask
#         -->temp3
#         annulus_mask has CREATOR & DATE attributes

    $filterexp="'mask("."$temp3".",0,0,X,Y)'";

    if(system("farith infil1=$temp1 infil2=$posit_mask+1 outfil=$temp3 ops=MUL datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFarith","Cannot calculate posmask*annulusmask\n");
    }

    $msg="Filtering events and extracting \"positive\" spectrum...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

#
#    8.2) Extract positive spectrum -->spec_pos
#
    if(system("evselect table=$evtlist:EVENTS  expression=$filterexp withfilteredset=false keepfilteroutput=false destruct=false writedss=true updateexposure=true withspectrumset=true spectrumset=$spec_pos energycolumn=PI spectralbinsize=$specBin withspecranges=true specchannelmin=0 specchannelmax=$specChanMax")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrEvselect","Cannot create Positive spectrum\n");
    }
#
#     8.3) Obtain Solid Angle for positive spectra: count "1" pixels
#          in filter mask
#

    if(system("fimgstat infile=$temp3 threshlo=1 threshup=1 outfile=$flatimg clobber=yes >/dev/null")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFimgstat","Cannot run fimgstat to obtain num.of pixels>0\n");
    }

    unlink($flatimg);
    $SolidAngle_pos=`pget fimgstat num`;
    $SolidAngle_pos=$SolidAngle_pos*($inPixSize/$xColPixSize)**2;
    unlink($temp3);

#
#    8.4) Do annulus_mask*neg_mask=final neg_mask -->temp3
#

    if(system("farith infil1=$temp1 infil2=$negat_mask+1 outfil=$temp3 ops=MUL datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFarith","Cannot calculate negmask*annulusmask\n");
    }

    $msg="Filtering events and extracting \"negative\" spectrum...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

#
#    8.5) Extract negative spectrum -->spec_neg
#
    if(system("evselect table=$evtlist:EVENTS  expression=$filterexp withfilteredset=false keepfilteroutput=false destruct=false writedss=true updateexposure=true withspectrumset=true spectrumset=$spec_neg energycolumn=PI spectralbinsize=$specBin withspecranges=true specchannelmin=0 specchannelmax=$specChanMax")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrEvselect","Cannot create Negative spectrum\n");

    }
#
#    8.6) Obtain Solid Angle for negative spectra:
#         count "1" pixels in filter mask
#
    if(system("fimgstat infile=$temp3 threshlo=1 threshup=1 outfile=$flatimg clobber=yes >/dev/null")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFimgstat","Cannot run fimgstat to obtain num.of pixels>0");
    }
    unlink($flatimg);
    $SolidAngle_neg=`pget fimgstat num`;
    $SolidAngle_neg=$SolidAngle_neg*($inPixSize/$xColPixSize)**2;
    unlink($temp1);
    unlink($temp3);

#
# ------------------------------------------------------ 
#   9) Subtract SPEC_NEG spectrum from SPEC_POS spectrum: MATHPHA
#       Input:  SPEC_NEG_file_in_annulus
#               SPEC_POS_file_in_annulus
#
#       Output: DIFF_SPEC_file_in annulus
# ------------------------------------------------------ 

    $nspec=$jj+1;

#
#     9.1)Normalise both spectra to Positive spectrum solid angle
#


    $msg="Normalising both spectra to Positive spectrum solid angle...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

    if($SolidAngle_pos == 0. && $SolidAngle_neg > 0.){
      $expr="'0.0*".$spec_pos."-(".$spec_neg.")'";
      $SolidAngle=$SolidAngle_neg;
      SAS::warning("NullSpecPos","Positive spectrum has null extraction area for offaxis=$off_arcm[$jj] (spectrum number $nspec) \n");	  

    }elsif($SolidAngle_pos > 0. && $SolidAngle_neg == 0.){
      $expr="'".$spec_pos."-0.0*".$spec_neg."'";
      $SolidAngle=$SolidAngle_pos;
      SAS::warning("NullSpecNeg","Negative spectrum has null extraction area for offaxis=$off_arcm[$jj] (spectrum number $nspec) \n");	  

    }elsif($SolidAngle_pos <= 0. || $SolidAngle_neg <= 0.){
      $SolidAngle=0.0;
      $expr="'0.0*".$spec_pos."-".$spec_neg."*0.0'";
      SAS::warning("NullSpec","Positive and Negative spectrum have null extraction area for offaxis=$off_arcm[$jj] (spectrum number $nspec) \n");
    }else{	
      $factor=$SolidAngle_pos/$SolidAngle_neg;
      $factorFormat=sprintf "%.3f",$factor;
      $expr="'".$spec_pos."-(".$spec_neg."*".$factorFormat.")'";
      $SolidAngle=$SolidAngle_pos;
    }

    $SolidAngleFormat=sprintf "%.5e",$SolidAngle;



#
#      9.2) EXPOSURE update just in case evselect cannot do it
#           (MATHPHA cannot operate without this keyword in datasets)
#

    $dset[0]=$spec_pos;
    $dset[1]=$spec_neg;

    for($ids=0;$ids<2;$ids++){
      $exposds=0.000;
      $expostb=0.000;

      if(system("fkeypar fits=$dset[$ids]+0 key='EXPOSURE'")){
	clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
	clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
	SAS::error("ErrFkeypar","Cannot run fkeypar to read EXPOSURE from \"$dset[$ids]\"\n");
      }
      $existExposDs=`pget fkeypar exist`;
      chomp($existExposDs);
      if ($existExposDs =~ /yes/){
	$exposds=`pget fkeypar value`;
	chomp($exposds);
      }
      if(system("fkeypar fits=$dset[$ids]\[SPECTRUM\] key='EXPOSURE'")){
	clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
	clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
	SAS::error("ErrFkeypar","Cannot run fkeypar to read EXPOSURE from \"$dset[$ids]\"\n");
      }
      $existExposTb=`pget fkeypar exist`;
      chomp($existExposTb);
      if($existExposTb =~ /yes/){
	$expostb=`pget fkeypar value`;
	chomp($expostb);
      }

      if($existExposTb =~ /no/){
	if($existExposDs =~ /no/){
	  $expostb=1.000;
	  SAS::warning("NullExpos1","Exposure value for spectrum $dset[$ids] cannot be properly updated by evselect in fluctuations spectrum number $nspec: Exposure updated to 1.000 \n");	  
	} 
	if(system("fparkey add=yes value=$expostb fits=$dset[$ids]\[SPECTRUM\] key='EXPOSURE'")){
	  clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
	  clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
	  SAS::error("ErrFparkey","Cannot run fparkey to add EXPOSURE to \"$dset[$ids]\"\n");
	}
      }

      if($expostb==0.0){
	if($exposds != 0.0){
	  $expostb=$exposds;
	  SAS::warning("NullExposDs","Exposure value for spectrum $dset[$ids] cannot be properly updated by evselect in fluctuations spectrum number $nspec: Exposure updated to EXPOSURE value in event list dataset \n");	  
	}else{
	  SAS::warning("NullExpos1","Exposure value for spectrum $dset[$ids] cannot be properly updated by evselect in fluctuations spectrum number $nspec: Exposure updated to 1.000 \n");	  
	  $expostb=1.0000;
	}
	if(system("fparkey value=$expostb fits=$dset[$ids]\[SPECTRUM\] key='EXPOSURE'")){
	  clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
	  clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
	  SAS::error("ErrFparkey","Cannot run fparkey to add EXPOSURE to \"$dset[$ids]\"\n");
	}
      }
    }

#
#       9.3) Use MATHPHA to calculate the difference spectrum
#

    if($ENV{"SAS_VERBOSITY"}){
      $verb=$ENV{"SAS_VERBOSITY"}-1;
    }else{
      $verb=-1;
    }

    $msg="Calculating difference spectrum...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

    $msg="Executing (routine):mathpha expr=$expr units='C' outfil=$temp3 exposure=$spec_pos Areascal='%' properr='yes' errmeth='POISS-1' ncomments=0 auxfiles='NONE' backfile='NONE' backscal=$SolidAngleFormat corrfile='NONE' corrscal='NONE' arfile='NONE' rmfile='NONE' phaversn='1.1.0' divzero=-99 chatter=$verb  clobber=yes\n";

    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);


    if(system("mathpha expr=$expr units='C' outfil=$temp3 exposure=$spec_pos Areascal='%' properr='yes' errmeth='POISS-1' ncomments=0 auxfiles='NONE' backfile='NONE' backscal=$SolidAngleFormat corrfile='NONE' corrscal='NONE' arfile='NONE' rmfile='NONE' phaversn='1.1.0' divzero=-99 chatter=$verb  clobber=yes")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrMathpha","MATHPHA: Subtracted spectrum cannot be created\n");
    }

    # remove CREATOR attribute


    if(system("fparkey value=0 fitsfile=$temp3+1 keyword=-CREATOR")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFparkey","Cannot run fparkey to remove CREATOR attribute");
    }

    (-e $spec_neg) && unlink($spec_neg);
    (-e $spec_pos) && unlink($spec_pos);

#
#      9.4) Add difference spectrum to final spectrum (new row):SPECADD
#

    $msg="Adding difference spectrum as a new row in output spectral file...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

    if($nspec == 1){

      if(system("specadd sets=$temp3 sumset=$flspecfile")){
	clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
	clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
	SAS::error("ErrFladdspec","Cannot store single spectrum\n");
      }
    }else{
      # rename multiple file to temp file to store one more single spectrum
      rename($flspecfile,$spec_pos);

      if(system("specadd sets=$temp3 oldsumset=$spec_pos sumset=$flspecfile")){
	clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
	clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
	SAS::error("ErrFladdspec","Cannot store single spectrum\n");
      }
    }
    (-e $spec_pos)&& unlink($spec_pos);
    (-e $temp3)&& unlink($temp3);
  } #end offaxis values

#------------------------------------------------------
#   10) Save offaxis values in ROWID column of multiple spectra
#------------------------------------------------------


  if(system("fcolpar infile=$flspecfile colname='ROWID' ")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFcolpar","Cannot establish column number for ROWID in fluxtuations file\n");
  }

  for($jj=1;$jj<=$nann;$jj++){
    $rowidStr=sprintf "%.5f",$off_arcm[$jj-1];
    $rowidStr .= "   arcm";
    if(system("fpartab value=\"$rowidStr\" fits=$flspecfile column='ROWID' row=$jj")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFpartab","Cannot run to fill in ROWID in \"$flspecfile\" in step $jj\n");
    }
  }


#------------------------------------------------------
#   11) Add some attributes
#------------------------------------------------------
#
  $label="Epic Fluctuations Spectra at 10 different off-axis positions specified in ROWID column";
  if(system("fparkey add=yes value=\"EPIC FLUCTUATIONS SPECTRUM\"  fits=$flspecfile+0  keyw='CONTENT' comm='$label'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFparkey","Cannot run fparkey to add CONTENT to \"$flspecfile\"\n");
  }

  $label="Multiple PHA file contained";
  if(system("fparkey add=yes value=\"COUNT\"  fits=$flspecfile+1  keyw='HDUCLAS3' comm='$label'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFparkey","Cannot run fparkey to add HDUCLAS3 to \"$flspecfile\"\n");
  }

# ------------------------------------------------------
#   12) Remove intermediate files
# ------------------------------------------------------
#
  clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
  clearTemporary($spec_pos,$spec_neg,$temp3);

}

sub clearTemporary{
  foreach $_ (@_){
    (-e $_) && unlink($_);
  }
}
